/*----------------------------------------------------------------------*/
/* Copyright (c) 1988-1989						*/
/* by CompuServe Inc., Tucson, AZ.  All Rights Reserved			*/
/* TIFRLE.C can be copied and distributed freely for any		*/
/* non-commercial purposes. TIFRLE.C can only be incorporated		*/
/* into commercial software with the permission of CompuServe Inc.	*/
/*----------------------------------------------------------------------*/

/* TIFRLE.C */

/*
 * ABSTRACT:
 *	This file contains routines to convert one-dimensional group-3 FAX
 *	compressed lines into Run Length Encoded lines.
 *
 *	See TIFRLE.H for a summary of functions.
 *
 * AUTHOR: Doug Chinnock
 *
 * REVISION HISTORY:
 *
 */

#include        <stdio.h>
#include	<string.h>

#include	"GIFbase.h"
#include	"cnvhuf.h"
#include	"tifrle.h"

extern	int	scan_lines; 

/* Define all the prototypes */

#include	"readtif.h"

/* Table of functions */

static void process_color( rle_line	*line_record,
			   int		run_length );


/* The two constant tables have the form:
|
|	table_1st_bits	convert_1st_bits; 
|
|   typedef
|	table_end_bits	INTER_ARR3[16];
|   INTER_ARR3		array_convert_end_bits; 
*/
#include	"huftab.h"

/* Top scope variables: */

static unsigned char
	head_byte, 	/* 8 bits at head of stream */
	used_bits; 	/* Bits needed to fill head_byte */

static rle_line
	current_line;	/* Holding buffer for current line */
static int
	just_eol;	/* Last line had no data */
static unsigned char
	nil_lines;	/* Count of just_eols */

static void process_color( rle_line	*line_record,
			   int		run_length )
{
if ((*line_record).rle_count < LONGEST_RLE)
   {
   ++ (*line_record).rle_count;
   (*line_record).rle_list[((*line_record).rle_count) - 1] = run_length;
   }

(*line_record).line_length += run_length;
} /* Process_color */

extern void 	expand_huffman_code( colors	color_of_run,
				     int	*length_of_color,
				     int	*eol_seen )
{
   int
	run_incr; 	/* Incremental run length */
   node_label
	last_node; 	/* Decode array for last 4 bits */

   /* Start of expand_huffman_code: */
   /* Initialize for the code */
   
   (*eol_seen) = 0;
   (*length_of_color) = 0;

   /* Now process a code */
   do			/* UNTIL ( eol_seen ) OR ( run_incr < 64 ) */
      {
      refill_byte(&(head_byte),used_bits);	/* Make sure 8 bits ready */
      used_bits = convert_1st_bits[head_byte][color_of_run].bits_consumed;

      if (used_bits == 0)		/* If over 8 bit code */
         {				/* need to use 4 bit extension */
         last_node = convert_1st_bits[head_byte][color_of_run].run_length;
         refill_byte(&(head_byte),8);	/* Get 8 more bits */
         used_bits = array_convert_end_bits[(last_node)-1][head_byte / 16]
            			.bits_consumed;
         run_incr = array_convert_end_bits[(last_node)-1][head_byte / 16]
            			.run_length;
         switch ((int)(used_bits)) {
            case (int)0:		/* EOL */
               {
               refill_byte(&(head_byte),4);	/* Consume EOL */
               (*eol_seen) = 1;			/* Indicate same */
               break;
               }
            case (int)1:
            case (int)2:
            case (int)3:
            case (int)4:
               /* 9, 10, 11, 12 bit codes */
               {
               break;
               }

            case (int)5:		/* All zeros means Fill */
               {
               do
                  {				/* Scan for EOL bit */
                  refill_byte(&(head_byte),1);
                  }
               while (head_byte < 16);

               refill_byte(&(head_byte),4);	/* Consume EOL */
               (*eol_seen) = 1;			/* Indicate same */
               used_bits = 0;			/* Don't consume next time in */
               break;
               }
            case (int)6:		/* Strange run of zeros */
               {
	       printf( "%s%d\n", "Confusing fill on line: ", scan_lines + 1);
               used_bits = 4;		/* Consume 4 bits next time in */
               break;
               }
            case (int)7:		/* 13 bit make up code */
               {
               if (((head_byte & 8)) == 8)
                  run_incr += 64;
               refill_byte(&(head_byte),5);
               used_bits = 0;			/* Don't consume next time in */
               break;
               }
            } /* Switch (int)(used_bits) */
         } /* IF used_bits = 0 */
      else /* IF used_bits <> 0 */
         {
         run_incr = convert_1st_bits[head_byte][color_of_run].run_length;
         }

      (*length_of_color) += run_incr; /* Accumulate total */
      }
   while ( !(*eol_seen) && (run_incr >= 64) );
} /* expand_huffman_code */

extern void	expand_huffman_line( rle_line	*emit_line,
				     int	*eol_first )
{
   colors
	color; 		/* Current mode */
   int
	color_length,	/* Length of this color's run */
	eol_encountered;/* Huffman expander saw eol code */

   /* Expand_huffman_line */
   
   (*emit_line).line_length = 0;
   (*emit_line).rle_count = 0;
   (*emit_line).rle_pointer = 0;
   (*emit_line).rle_color = WHITE;
   (*emit_line).rle_start_column = 0;
   (*emit_line).current_column = 0;

   (*eol_first) = 1;		/* Preset to eol before data */

   /* Now process a line */
   if ( nil_lines < 6 )			/* If not past end of page */
      {
      color = WHITE;

      do
	 {
	 expand_huffman_code(color,&(color_length),&(eol_encountered));
	 if (!(eol_encountered)) (*eol_first) = 0;
	 if (color_length >= 0) process_color(&((*emit_line)), color_length);
	 color = !(color);		/* Change color */
	 }
      while (!(eol_encountered));
   
      ++ scan_lines;
      };

/* debug info *//*
 printf( "%s%d%s%d\n",
	 "Line ",
	 scan_lines,
	 ", length = ",
	 (*emit_line).line_length );
/* debug info */

   if ((*emit_line).line_length > LONGEST_LINE)
      printf("%s%d%s%d\n",
             "Line ", scan_lines,
             "OVERFLOW; length= ", (*emit_line).line_length );
} /* Expand_huffman_line */

extern short	get_RLE_bit()
{
   if ( current_line.current_column >= FAX_line_width )
      {			/* Line has been fully sent */
			/* get another line in */
      expand_huffman_line( &(current_line), &(just_eol) );

      if (just_eol)
        {
	nil_lines ++;
	if ( nil_lines >= 6 ) return -1;	/* Detect EOF */
	}
      else
	nil_lines = 0;

      }; /* if ( current_line.current_column >= FAX_line_width ) */

   if ( current_line.current_column
        < current_line.line_length )
      {				/* Have more pixels in RLE stream */
      if ( (current_line.current_column - current_line.rle_start_column)
	   < current_line.rle_list[ current_line.rle_pointer ] )
	 {			/* Have more pixels in this run length */
	 current_line.current_column ++;	/* Bump column */
	 return current_line.rle_color;
	 }
      else
         {			/* Have used up all pixels in this run */
         current_line.rle_color = ! current_line.rle_color;
         current_line.rle_pointer ++;		/* Bump to next RL */
         current_line.rle_start_column = current_line.current_column;
	 current_line.current_column ++;	/* Bump column */
	 return current_line.rle_color;
         }
      }
   else /* ( current_line.current_column >= current_line.line_length ) */
      {				/* RLE stream has run out of data */
      if ( current_line.current_column < FAX_line_width )
	 {			/* Need to finish off line */
	 current_line.current_column ++;	/* Bump column */
	 return WHITE;
	 }
      }; /* if ( current_line.current_column >= current_line.line_length ) */
} /* get_RLE_bit */

int
	scan_lines = 0;		/* Initialize to establish location */

extern short int	start_huffman_page( char huffman_name[] )
{
   scan_lines = 0;
   if ( initialize_stream(huffman_name) != 0 )
	return -1;

   /* Read in first line of Huffman image */

   nil_lines = 0;
   used_bits = 8;
   expand_huffman_line( &(current_line), &(just_eol) );
   return 0;
}
